function [fEvals, exitflag, rec] = AS_ES(f, g, fOpt, delta, epsilon, maxSteps, x0, sigma0)

% [fEvals, exitflag, rec] = AS_ES(f, g, fOpt, delta, epsilon, maxSteps, x0, sigma0)
%
% (1+1)-Active-Set-Evolution Strategy for solving explicitly constrained
% black-box optimization problems.  See
%
% Patrick Spettel, Zehao Ba, and Dirk V. Arnold,
% Active sets for explicitly constrained evolutionary optimization,
% submitted, 2020.
%
% for details.
%
%   f is the objective function
%
%   g are the constraint functions (in the format required by fmincon)
%
%   fOpt is the objective function value of the optimal solution to the
%     problem of minimizing f subject to constraints g
%
%   delta is the constraint tolerance
%
%   epsilon is the target accuracy
%
%   maxSteps is the maximum number of iterations
%
%   x0 and sigma0 (both optional) are starting point and starting step size
%
%   fEvals is the number of objective function evaluations performed
%
%   exitflag indicates whether or not the run was successful
%
%   rec is a detailed record of the entire run
%
%  Example:
%    [fEvals, exitflag, rec] = AS_ES(@f01, @g01, -15, 1.0e-08, 1.0e-08, 2000);
%
%  (c) Dirk V. Arnold, 2020.
%  Permission to use or modify this code for any purpose is granted,
%  provided that this message is left intact.

  [nonlcon, A, b, Aeq, beq, lb, ub] = g();
  n = size(A, 2);               % number of variables
  l = size(nonlcon(lb), 1);		% number of nonlinear inequality constraints
  m = size(b, 1);               % number of linear inequality constraints
  I = eye(n);
  
  if nargin>6                   % starting point given
    sigma = sigma0;
    [x, exitflag, AS] = project(x0, false(l+m+2*n, 1));
    if exitflag<1
      disp 'cannot generate a feasible starting point'
      exit
    end
  else                          % find a feasible initial solution
    if isempty(lb) || isempty(ub)
      disp 'need to supply either bounds or a starting point'
      exit
    end
    sigma = 0.2*min(ub-lb);
    exitflag = 0;
    while exitflag<1
      x = lb + (ub-lb).*rand(n, 1);
      [x, exitflag, AS] = project(x, false(l+m+2*n, 1));
    end
  end

  lastChecked = zeros(l+m+2*n, 1);
  fx = f(x);
  fEvals = 1;
  if isnan(fx) || ~isreal(fx)
    fx = inf;
  end

  rec = zeros(maxSteps, 4+3*n+m+l);
  t = 1;
  rec(t, :) = [fEvals, n, AS', x', fx-fOpt, sigma];

  while t<maxSteps && fx-fOpt>abs(fOpt)*epsilon && sigma>1.0e-08

    [~, ~, nc, nceq] = nonlcon(x);
    N = [nc(:, AS(1:l)), nceq, A(AS(l+1:l+m), :)', Aeq', ...
              -I(:, AS(l+m+1:l+m+n)), I(:, AS(l+m+n+1:end))];
    idx = any(~isfinite(N));
    nEff = n-rank(N(:, ~idx))-sum(idx);

    as = AS;
    if nEff==0 || (any(as) && rand<0.2)
      [~, idx] = min(lastChecked+rand(size(lastChecked))+(1+max(lastChecked)).*~as);
      as(idx) = false;
    else
      idx = 0;
    end
    
    % generate offspring
    k = 0;
    while k<400
      [y, exitflag, tight] = project(x+sigma*randn(n, 1), as);
      tight = as | tight;
      if exitflag>0 && (idx==0 || ~tight(idx))
        if idx>0
          lastChecked(idx) = t;
        end
        break
      end
      k = k+1;
    end
    
    if k==400         % failure to generate a viable offspring
     if idx>0
        if rand<0.2
          AS(idx) = false;
        elseif any(as)
         lastChecked(idx) = min(lastChecked(as));
        end
     end
      continue
    end

    % selection
    fy = f(y);
    fEvals = fEvals+1;

    if fy<fx                % success
      x = y;
      fx = fy;
      AS = tight;
      sigma = sigma*exp(~any(lastChecked==t)*1.0/sqrt(1+nEff));
    else                    % failure
      sigma = sigma*exp(~any(lastChecked==t)*(-0.25)/sqrt(1+nEff));
    end

    t = t+1;
    rec(t, :) = [fEvals, nEff, AS', x', fx-fOpt, sigma];

  end

  rec = rec(1:t, :);
  exitflag = fx-fOpt<=abs(fOpt)*epsilon;

  % -- project ------------------------------------------------------------

  function [y, exitflag, tight] = project(x, AS)
  % find the feasible solution closest to x
  
    function [ASc, ASceq, ASnc, ASnceq] = ASnonlcon(x)
    % make active inequality constraints equality ones
      [c, ceq, nc, nceq] = nonlcon(x);
      ASc = c(~AS(1:l));
      ASceq = [ceq; c(AS(1:l))];
      ASnc = nc(:, ~AS(1:l));
      ASnceq = [nceq, nc(:, AS(1:l))];
    end
    
    % assemble options for fmincon
    opts(1) = optimoptions(@fmincon, ...
                        'Algorithm', 'active-set', ...
                		'Display', 'off', ...
                        'OutputFcn', @outfun, ...
                        'SpecifyConstraintGradient', true, ...
                        'ConstraintTolerance', delta);
    opts(2) = optimoptions(@fmincon, ...
                        'Algorithm', 'sqp', ...
                		'Display', 'off', ...
                        'SpecifyConstraintGradient', true, ...
                        'ConstraintTolerance', delta);

    ASAeq = [Aeq; A(AS(l+1:l+m), :); -I(AS(l+m+1:l+m+n), :); I(AS(l+m+n+1:end), :)];
    ASbeq = [beq; b(AS(l+1:l+m)); -lb(AS(l+m+1:l+m+n)); ub(AS(l+m+n+1:end))];

    % enforce the constraints
    kk = 1;
    while kk<=length(opts)
      try
        [y, ~, exitflag, ~, lambda] = fmincon(@(y) sum((y-x).^2), x, ...
              A(~AS(l+1:l+m), :), b(~AS(l+1:l+m)), ASAeq, ASbeq, lb, ub, ...
              @ASnonlcon, opts(kk));
        [c, ceq] = ASnonlcon(y);
        maxVio = max([c; abs(ceq); A*y-b; abs(ASAeq*y-ASbeq); lb-y; y-ub]);
      catch MException  % fmincon crash
        y = [];
        exitflag = -27;
      end
      
      if opts(kk).SpecifyConstraintGradient && (exitflag==-1 || exitflag==-27)
        opts(kk).SpecifyConstraintGradient = false;
        continue
      end

      if exitflag>=1 && isreal(y) && maxVio<delta
        tmp = zeros(l, 1);
        tmp(~AS(1:l)) = lambda.ineqnonlin;
        lambda.ineqnonlin = tmp;

        tmp = zeros(m, 1);
        tmp(~AS(l+1:l+m)) = lambda.ineqlin;
        lambda.ineqlin = tmp;

        exitflag = 1;
        tight = [lambda.ineqnonlin; lambda.ineqlin; lambda.lower; lambda.upper]>0.0 ...
                  & abs([nonlcon(y); A*y-b; lb-y; y-ub])<delta;
        return
      end
      kk = kk+1;
    end
    exitflag = 0;
    tight = false(l+m+2*n, 1);
  end

  function stop = outfun(x, optimValues, state)  % avoid infinite loops in fmincon
    [~, ~, nc, nceq] = nonlcon(x);
    stop = ~isreal(nc) || any(~isfinite(nc(:))) || ~isreal(nceq) || any(~isfinite(nceq(:)));
  end

end